/*
 * Copyright 2002 Oskar Ojala and Henri Sivonen.
 *
 * The contents of this file are subject to the Artistic License.
 * You may not use this file except in compliance with the Artistic
 * License (which you should have received along with this file).
 */

package dualroids;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.font.*;
import java.awt.image.*;

import java.io.*;
import java.lang.*;

import javax.swing.*;

/**
 * Draws the actual playing field.
 * Takes care of all actual drawing and buffering
 * during the game.
 *
 * @author Oskar Ojala, omojala@cc.hut.fi
 * @author Henri Sivonen, hsivonen@iki.fi
 */

final class SpaceArea extends JComponent {

    /** Width of the displayable area. Must match <code>Arena.HEIGHT</code> */
    public static final int WINDOW_WIDTH = 640;

    /** Height of the displayable area. Must match <code>Arena.WIDTH</code> */
    public static final int WINDOW_HEIGHT = 480;

    /** File name of the background image */
    private static final String BG_GRAPHIC = "dualroids/stars.jpg";

    /** Font for displaying scoreboard during the game */
    private final static Font SCORE_BOARD_FONT = new Font("Helvetica", 0, 16);

    /** Font for displaying high scores when the game ends */
    private final static Font SCORE_FONT = new Font("Helvetica", 0, 24);

    /** Font for displaying the 'Other player has quit' text */
    private final static Font PLAYER_HAS_QUIT_FONT = new Font("Helvetica", 0, 46);

    /** Font for displaying the 'Game Over' -text */
    private final static Font GAME_OVER_FONT = new Font("Helvetica", 0, 72);

    /** A 'clear' color, fully transparent */
    private final static Color CLEAR = new Color(255,255,255,0);

    /** A semi-transparent gray color for showing end texts */
    private final static Color BLEAK_GRAY = new Color(255,255,255,190);
    
    /** The connector to the client */
    private ClientModelConnector connector;

    /** Name of player 1 */
    private String player1 = "Player 1";

    /** Name of player 2 */
    private String player2 = "Player 2";

    /** Score of player 1 */
    private String score1 = "0";

    /** Score of player 2 */
    private String score2 = "0";

    /** Number of ships player 1 has left */
    private String ships1 = "5";

    /** Number of ships player 2 has left */
    private String ships2 = "5";

    /** The background image */
    private Image bgImage;

    /** Toggles whether the score board has been updated */
    private boolean scoreboardUpdated;

    /** Toggles whether other player has quit */
    private boolean otherPlayerQuit;

    /** Toggles whether the game is over */
    private boolean gameOver;

    /** Rendering hints for rendering asteroids, spaceships etc. */
    private final RenderingHints hints;

    /** Background image buffer */
    private BufferedImage buffer1;

    /** Foreground image buffer, works best with accelerated graphics */
    private BufferedImage buffer2;

    /** An arena object, which is used to fetch the shapes to draw*/
    private Arena arena;

    /** A high score-handler to save/load highscores */
    private HighScoreHandler scoreHandler;

    /**
     * Registers an <code>Arena</code> for painting.
     * @param arena the <code>Arena</code> to be painted
     */
    public void registerArena(Arena arena) {
        this.arena = arena;
    }
    
    /**
     * Constructor for this class: initializes all necessary parameters.
     *
     * @param cmc The ClientModelConnector for the client whose screen
     * this is displayed on.
     */
    public SpaceArea(ClientModelConnector cmc) {

        super();
        this.bgImage = (new ImageIcon(BG_GRAPHIC)).getImage();
        this.connector = cmc;
        this.setOpaque(true);

        //some window managers only require setPreferredSize()
        this.setMinimumSize(new Dimension(WINDOW_WIDTH,WINDOW_HEIGHT));
        this.setMaximumSize(new Dimension(WINDOW_WIDTH,WINDOW_HEIGHT));
        this.setPreferredSize(new Dimension(WINDOW_WIDTH,WINDOW_HEIGHT));

        this.hints = new RenderingHints(null);
        this.hints.put(RenderingHints.KEY_ANTIALIASING,
                       RenderingHints.VALUE_ANTIALIAS_ON);
        this.hints.put(RenderingHints.KEY_DITHERING,
                       RenderingHints.VALUE_DITHER_DISABLE);
        this.hints.put(RenderingHints.KEY_INTERPOLATION,
                       RenderingHints.VALUE_INTERPOLATION_NEAREST_NEIGHBOR);
        this.hints.put(RenderingHints.KEY_RENDERING,
                       RenderingHints.VALUE_RENDER_SPEED);
        this.hints.put(RenderingHints.KEY_COLOR_RENDERING,
                       RenderingHints.VALUE_COLOR_RENDER_SPEED);
        // antialiasing is set in Runner
        this.hints.put(RenderingHints.KEY_TEXT_ANTIALIASING,
                       Runner.TEXT_ANTIALIASING);

        //on some platforms, a buffer the same size as the component
        //doesn't cover the entire window, thus a little extra added 
        //to the size
        buffer1 = new BufferedImage(WINDOW_WIDTH+5, WINDOW_HEIGHT+5, 
                                    BufferedImage.TYPE_INT_RGB);

        //used with fast computer/accelerated graphics
        /*
        buffer2 = new BufferedImage(WINDOW_WIDTH, WINDOW_HEIGHT, 
                                    BufferedImage.TYPE_INT_ARGB);
        */

        //insures that proper texts are drawn when the game begins
        this.scoreboardUpdated = true;
        this.otherPlayerQuit = false;
        this.gameOver = false;
    }

    /**
     * Called externally when the other player has quit.
     * Sets the toggles so the appropriate texts are drawn.
     */
    public final void otherPlayerQuit() {
        synchronized(this) {
            this.otherPlayerQuit = true;
            this.scoreboardUpdated = true;
        }
    }

    /**
     * Called externally when the game is over.
     * Sets the toggles so the appropriate texts are drawn.
     */
    public final void gameOver() {
        synchronized(this) {
            this.gameOver = true;
        }
    }

    /**
     * Updates player names, number of ships and current score.
     * Sets the toggles so the scoreboard is drawn.
     *
     * @param player1 The name of player 1
     * @param player2 The name of player 2
     * @param score1 The score of player 1
     * @param score2 The score of player 2
     * @param ships1 Number of ships player 1 has remaining
     * @param ships2 Number of ships player 2 has remaining
     */
    public final void updateScoreBoard(String player1,
                                       String player2,
                                       String score1,
                                       String score2,
                                       String ships1,
                                       String ships2)
    {
        synchronized(this) {         
            this.player1 = player1;
            this.player2 = player2;
            this.score1 = score1;
            this.score2 = score2;
            this.ships1 = ships1;
            this.ships2 = ships2;

            this.scoreboardUpdated = true;
        }
    }
    
    /**
     * Draws end credits when the game ends.
     * consumes memory far faster than can be freed,
     * but as the game has ended anyway, it doesn't
     * matter much.
     *
     * @param g2_b2 The graphics context which to paint into
     */
    public void drawEndCredits(Graphics2D g2_b2) {
        
        g2_b2.setRenderingHints(this.hints);
        g2_b2.setPaint(BLEAK_GRAY);

        if ( gameOver ) {
            
            //draws "Game Over" text
            TextLayout gameOverTL = 
                new TextLayout("Game Over", GAME_OVER_FONT, 
                               g2_b2.getFontRenderContext());
            Rectangle2D bounds = gameOverTL.getBounds();
            int width = (int)bounds.getWidth();
            int height = (int)bounds.getHeight();
            gameOverTL.draw(g2_b2, (WINDOW_WIDTH-width)/2, 
                            (WINDOW_HEIGHT-height)/2 - 35);
            
        } else if (otherPlayerQuit) {
            
            //draws "other player has quit" text
            TextLayout otherPlayerQuitTL = 
                new TextLayout("Other player has quit", PLAYER_HAS_QUIT_FONT, 
                               g2_b2.getFontRenderContext());
            Rectangle2D bounds = otherPlayerQuitTL.getBounds();
            int width = (int)bounds.getWidth();
            int height = (int)bounds.getHeight();
            otherPlayerQuitTL.draw(g2_b2, (WINDOW_WIDTH-width)/2, 
                                  (WINDOW_HEIGHT-height)/2 - 35);
        }

        //fetches the highscores
        if(this.scoreHandler == null) {
            this.scoreHandler = new HighScoreHandler(connector);
        }
        String[] names = this.scoreHandler.getNames();
        int[] points = this.scoreHandler.getPoints();

        TextLayout namesTL, scoresTL;

        //draws the highscore list
        for (int i = 0; i < 5; i++) {
            namesTL = 
                new TextLayout(names[i], SCORE_FONT, 
                               g2_b2.getFontRenderContext());
            scoresTL = 
                new TextLayout(("" + points[i]), SCORE_FONT, 
                               g2_b2.getFontRenderContext());
            Rectangle2D bounds_n = namesTL.getBounds();
            Rectangle2D bounds_s = scoresTL.getBounds();
            int height = (int)bounds_s.getHeight() + 8;
            namesTL.draw(g2_b2, 220, 
                         (WINDOW_HEIGHT-5*height)/2 + i*height + 60);
            scoresTL.draw(g2_b2, 390, 
                          (WINDOW_HEIGHT-5*height)/2 + i*height + 60);
        }
    }

    /**
     * Paints scoreboard into buffer1 with the background
     * image. The scores will thus be under the asteroids,
     * but on slower machines it's noticeably faster than
     * using drawScoreBoard().
     */
    public void paintScoreBoard() {
        Graphics2D g2_b1 = buffer1.createGraphics();
        
        g2_b1.setRenderingHints(this.hints);

        g2_b1.setPaint(Color.black);
        g2_b1.setFont(SCORE_BOARD_FONT);
        //on some platforms, a buffer the same size as the component
        //doesn't cover the entire window, thus a little extra added 
        //to the size
        g2_b1.fillRect(0, 0, WINDOW_WIDTH+10, WINDOW_HEIGHT+10);
        g2_b1.drawImage(this.bgImage, new AffineTransform(), null);

        g2_b1.setPaint(Color.white);
        g2_b1.drawString(player1, 40, 30);
        g2_b1.drawString(player2, 530, 30);

        g2_b1.drawString(score1, 40, 50);
        g2_b1.drawString(score2, 530, 50);
        g2_b1.drawString("Ships left: " + ships1, 40, 450);
        g2_b1.drawString("Ships left: " + ships2, 530, 450);

        g2_b1.dispose();
    }

    /**
     * Draws the scoreboard into buffer2, so that scores can
     * float above the asteroids (buffer2 has an alpha channel).
     * On some machines, it slows redrawing down considerably.
     */
    public void drawScoreBoard() {
        Graphics2D g2_b2 = buffer2.createGraphics();
        g2_b2.setRenderingHints(this.hints);
        
        g2_b2.setBackground(CLEAR);
        g2_b2.clearRect(0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);

        g2_b2.setPaint(Color.white);
        g2_b2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                               RenderingHints.VALUE_ANTIALIAS_OFF);

        g2_b2.drawString(player1, 40, 30);
        g2_b2.drawString(player2, 530, 30);

        g2_b2.drawString(score1, 40, 50);
        g2_b2.drawString(score2, 530, 50);
        g2_b2.drawString("Ships left: " + ships1, 40, 450);
        g2_b2.drawString("Ships left: " + ships2, 530, 450);

        g2_b2.dispose();
    }

    /**
     * Draws the game screen and updates it.
     *
     * @param g The graphics context which to paint into
     */
    public void paint(Graphics g) {

        Graphics2D g2 = (Graphics2D) g;
        g2.setRenderingHints(this.hints);

        //checks whether scoreboard needs to be redrawn
        if (this.scoreboardUpdated) {
            this.paintScoreBoard();
            this.scoreboardUpdated = false;
        }
        //draws background and scores
        g2.drawImage(buffer1, 0, 0, this);
                
        g2.setStroke(new BasicStroke(2.0f));

        //draw asteroids, ships and bullets
        long time = this.connector.currentServerMillis();
        if(this.arena != null) {
            Shape currentShape;
            SpaceShip ship;
            Asteroid[] asteroids = arena.getAsteroids();
            SpaceShip[] ships = arena.getSpaceShips();
            Bullet[] bullets;
            for(int i = 0; i < ships.length; i++) {
                ship = ships[i];
                bullets = ship.getBullets();
                g2.setPaint(Color.white);
                for(int j = 0; j < bullets.length; j++) {
                    currentShape = bullets[j].getTransformedShape(time);
                    if(currentShape != null) {
                        g2.draw(currentShape);
                    }
                }
                if(i == this.connector.getShipIndex()) {
                    ship.updateState(time);
                    g2.setPaint(Color.white);
                } else {
                    g2.setPaint(Color.black);
                }
                currentShape = ship.getTransformedShape(time);
                if(currentShape != null) {
                    g2.fill(currentShape);
                    g2.setPaint(ship.getOutlineColor());
                    g2.draw(currentShape);
                }   
            }
            for(int i = 0; i < asteroids.length; i++) {
                currentShape = asteroids[i].getTransformedShape(time);
                if(currentShape != null) {
                    g2.setPaint(Color.black);
                    g2.fill(currentShape);
                    g2.setPaint(asteroids[i].getOutlineColor());
                    g2.draw(currentShape);
                }
            }
        }

        //used if you want the scoreboard to float over the asteroids
//         if (this.scoreboardUpdated) {
//             this.drawScoreBoard();
//             this.scoreboardUpdated = false;
//         }
//         g2.drawImage(buffer2, 0, 0, this);


        if (gameOver || otherPlayerQuit)
            drawEndCredits(g2);

        repaint();
    }
}
